# (C) 2018 - TECHNISCHE UNIVERSITEIT EINDHOVEN
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED,INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE, TITLE AND NON-INFRINGEMENT. IN NO EVENT SHALL THE
# COPYRIGHT HOLDERS OR ANYONE DISTRIBUTING THE SOFTWARE BE LIABLE FOR ANY DAMAGES
# OR OTHER LIABILITY, WHETHER IN CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT
# OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# THIS SOFTWARE IS PART OF THE TU/e COMPUTER SCIENCE REPORT 2018-01
# Citation for published version:
#
# Sanchez, L., Wesselink, J.W., & Willemse, T.A.C. (2018). BDD-based parity game solving: a comparison of
# Zielonka's recursive algorithm, priority promotion and fixpoint iteration. (Computer science reports; Vol. 1801).
# Eindhoven: Technische Universiteit Eindhoven.
# https://pure.tue.nl/ws/files/92755535/CSR_18_01.pdf

import configBDD
import solver_library
import generate_random_game
import read_game
from time import process_time
import sys, os, random, math
import subprocess

def Read_PBES(name, debug=False):
  # HIDEOUS AD-HOC PARSER FOR THE OUTPUT OF pbes2bdd
  # COMES WITH NO GUARANTEES
  result = subprocess.run(['pbes2bdd',name], stdout=subprocess.PIPE).stdout.decode('ISO-8859-1')
  l = result. split('--- edges ---')
  e = l[1].strip(). replace("'''",""). replace("\n","")
  
  EDGES = str(e)
  if debug: print('Edges = %s' %(EDGES))
  l = l[0]. split('--- priorities ---')
  k = l[1]. strip()
  prio = {}
  for line in k. split('\n'):
    (v, s) = line. split('=')
    value = int(v[v.find('[')+1:v.find(']')])
    formula = str(s.strip())
    prio[value] = formula
  l = l[0]. split('--- odd nodes ---')
  ODD = l[1].strip()
  if debug: print('Odd = %s' %(ODD))
  l = l[0]. split('--- even nodes ---')
  EVEN = l[1].strip()
  if debug: print('Even = %s' %(EVEN))
  l = l[0]. split('--- all nodes ---')
  VERTICES = l[1].strip()
  if debug: print('Vertices = %s' %(VERTICES))
  l = l[0]. split('--- bdd variables ---')
  variables = [ str(variable.strip().replace("'",'')) for variable in l[1].strip().split(',')]

  (all_vars,variables_) = ([],[])
  for var in variables:
    variables_. append('%s_' %(var))
    all_vars.append('%s'%(var))
    all_vars.append('%s_'%(var))
  configBDD.bdd. declare(*all_vars)

  Priorities = {}
  for i in prio:
    if debug: print('prio[%d] = %s' %(i,prio[i]))
    e = configBDD.bdd.add_expr('True & %s' %(prio[i]))
    if e != configBDD.bdd.false: Priorities[i] = e
  #test coherence
  e = configBDD.bdd.false
  for i in Priorities:
    e = e | Priorities[i]
    for j in Priorities:
      if i != j:
        if Priorities[i] & Priorities[j] != configBDD.bdd.false: print('grmbl')
  #/test
  Vertices = configBDD.bdd.add_expr(VERTICES)
  if e != Vertices: print('lost some vertices?!')
  Even = configBDD.bdd.add_expr(EVEN)
  Odd = configBDD.bdd.add_expr(ODD)
  Edges = configBDD.bdd.add_expr(EDGES)
  return (variables, variables_, Vertices, Edges, Even, Odd, Priorities)


debug = False
if len(sys.argv) < 2:
  print('Usage: python3 solve_pbes.py <file.pbes> [solver]')
  print('       where solver is either PP, zielonka or fixpoint') 
  print('       and <file.pbes> contains a .pbes file meeting all sorts')
  print('       of requirements (too many to list) and possibly a solver as arguments')
  sys.exit()
print('Processing PBES '+sys.argv[1])
(variables, variables_, V, Edges, Even, Odd, Priorities) = Read_PBES(sys.argv[1],debug)
h = solver_library.PG (variables,variables_, V, Edges, Even, Odd, Priorities)
g = h.duplicate()
k = h.duplicate()
h. info ()
if len(sys.argv) > 2 and sys.argv[2] in ['PP','zielonka','fixpoint']:
  print('Solving using %s ' %(sys.argv[2]), end='')
  solver_start = process_time()
  (W0,W1) = eval('g. %s (debug)' %(sys.argv[2]))
  solver_end = process_time()
  print('took %.2f seconds' %(solver_end-solver_start))
else:
  print()
  k. compact () # we first must compact the game
  print('Solving using Fixpoint iteration ', end='')
  fixpoint_solver_start = process_time()
  (W0__,W1__) = k. fixpoint (debug)
  fixpoint_solver_end = process_time()
  print('took %.2f seconds' %(fixpoint_solver_end-fixpoint_solver_start))

  print()
  print('Solving using Zielonka ', end='')
  (W0,W1) = g. zielonka (debug)
  zielonka_solver_start = process_time()
  (W0,W1) = g. zielonka (debug)
  zielonka_solver_end = process_time()
  print('took %.2f seconds' %(zielonka_solver_end-zielonka_solver_start))
  
  print()
  print('Solving using PP ', end='')
  pp_solver_start = process_time()
  (W0_,W1_) = h.PP(debug)
  pp_solver_end = process_time()
  print('took %.2f seconds' %(pp_solver_end-pp_solver_start))

 
  # sanity checks
  if W1_ != g. attractor('odd', W1_):
    print('ERROR: unstable odd solution from PP... %s won according to algorithm but attractor = %s' %(configBDD.bdd.to_expr(W1_),configBDD.bdd.to_expr(g.attractor('odd',W1_) )))
  if W0_ != g.attractor('even',W0_):
    print('ERROR: unstable even solution from PP... %s won according to algorithm but attractor = %s' %(configBDD.bdd.to_expr(W0_),configBDD.bdd.to_expr(g.attractor('even',W0_) )))
  if W0 != W0_:
    print('hmmm...; according to PP, %d vertices are won by Even and %d by Odd' %(configBDD.bdd.count(W0_,len(g.variables)),configBDD.bdd.count(W1_,len(h.variables))))
    print('         according to Zielonka, %d vertices are won by Even and %d by Odd' %(configBDD.bdd.count(W0,len(g.variables)),configBDD.bdd.count(W1,len(g.variables))))
    sys. exit()
  if W0 != W0__:
    print('hmmm...; according to Fixpoint iteration, %d vertices are won by Even and %d by Odd' %(configBDD.bdd.count(W0__,len(k.variables)),configBDD.bdd.count(W1__,len(k.variables))))
    print('         according to Zielonka, %d vertices are won by Even and %d by Odd' %(configBDD.bdd.count(W0,len(g.variables)),configBDD.bdd.count(W1,len(g.variables))))
    sys. exit()
   
