# (C) 2018 - TECHNISCHE UNIVERSITEIT EINDHOVEN
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED,INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE, TITLE AND NON-INFRINGEMENT. IN NO EVENT SHALL THE
# COPYRIGHT HOLDERS OR ANYONE DISTRIBUTING THE SOFTWARE BE LIABLE FOR ANY DAMAGES
# OR OTHER LIABILITY, WHETHER IN CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT
# OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# THIS SOFTWARE IS PART OF THE TU/e COMPUTER SCIENCE REPORT 2018-01
# Citation for published version:
#
# Sanchez, L., Wesselink, J.W., & Willemse, T.A.C. (2018). BDD-based parity game solving: a comparison of
# Zielonka's recursive algorithm, priority promotion and fixpoint iteration. (Computer science reports; Vol. 1801).
# Eindhoven: Technische Universiteit Eindhoven.
# https://pure.tue.nl/ws/files/92755535/CSR_18_01.pdf

import configBDD
from dd import autoref as _bdd

class PG:
  def __init__ (self,variables, variables_, V, E, even, odd, p):
    self. variables = variables
    self. variables_ = variables_
    self. substitution_list = { variables[i] : variables_[i] for i in range(len(variables)) }
    self. reverse_substitution_list = { variables_[i] : variables[i] for i in range(len(variables)) }
    self. V = V
    self. E = E
    self. even = even
    self. odd = odd
    self. p = dict(p)

  def duplicate (self):
    h = PG(self.variables,self.variables_, self.V, self.E, self.even, self.odd, self.p)
    return h

  def compact (self):
    ## ensure priorities are dense
    # preprocess priorities; required for the fixpoint algorithm
    d_min = min(list( self. p))
    (p_, d) = ({},d_min %2)
    priorities = list(self.p)
    priorities. sort()
    parity_list = [ i%2 for i in priorities ]
    for i in range(len(priorities)):
      try: p_[d] = p_[d] | self.p[priorities[i]] 
      except: p_[d] = self.p[priorities[i]]
      if i+1 < len(priorities):
        d = d + 1 if parity_list[i] != parity_list[i+1] else d
    self.p = dict(p_) 

  def has_successor (self, U):
    #check whether U = { u in V | u in U /\ exists v in V: u --> v } 
    V_ = configBDD.bdd. let(self. substitution_list, self. V)
    U_next = configBDD.bdd. let(self. substitution_list, U)
    Z = configBDD.bdd. quantify(U_next & self. E, self. variables_, forall = False) & U
    return (Z == U)
    
  def successor (self, U):
    U_next = configBDD.bdd. quantify(self. E & U, self. variables, forall = False)
    U_next = U | configBDD.bdd. let(self.reverse_substitution_list, U_next)
    return U_next

  def reachable (self, U):
    U_next = self. successor(U)
    i = 0
    while U != U_next:
      U = U_next
      U_next = self. successor(U)
      i = i +1
      print('  at level %d' %(i))
    return U_next

  def predecessor (self, player, U):
    # U is a BDD representing a set of vertices
    # player is either string 'even' or string 'odd'
    (V_player,V_opponent) = (self.even, self.odd) if (player == 'even') else (self.odd, self.even)
    V_ = configBDD.bdd. let(self. substitution_list, self. V)
    U_next = configBDD.bdd. let(self. substitution_list, U)
    U_player = V_player & configBDD.bdd. quantify(U_next & self. E, self. variables_, forall = False)
    # V_opponent /\ {v in V | forall u in V: v --> u ==> u in U } = 
    # V_opponent /\ {v in V | ~ (exists u in V: v --> u /\ u in V\U) } 
    U_opponent =  V_opponent & ~(configBDD.bdd. quantify(self. E & V_ & ~U_next, self. variables_, forall = False) )
    # return union of the two sets
    return U_player | U_opponent
   
  def predecessor_gen (self, player, X, U):
    # X,U are BDDs representing a set of vertices
    # X is used to restrict the set of edges to stay within X
    # player is either string 'even' or string 'odd'
    (V_player,V_opponent) = (self.even, self.odd) if (player == 'even') else (self.odd, self.even)
    V_ = configBDD.bdd. let(self. substitution_list, self. V)
    X_ = configBDD.bdd. let(self. substitution_list, X)
    E  = self. E & X & X_
    U_next = configBDD.bdd. let(self. substitution_list, U)
    U_player = V_player & configBDD.bdd. quantify(U_next & E, self. variables_, forall = False)
    # V_opponent /\ {v in V | forall u in V: v --> u ==> u in U } = 
    # V_opponent /\ {v in V | ~ (exists u in V: v --> u /\ u in V\U) } 
    U_opponent =  V_opponent & ~(configBDD.bdd. quantify(E & V_ & ~U_next, self. variables_, forall = False) )
    # return union of the two sets
    return U_player | U_opponent
 
 
  def attractor (self, player, A):
    # U is a BDD representing a set of vertices
    # player is either string 'even' or string 'odd'
    # attractor computation is a least fixpoint computation
    tmp = configBDD.bdd. false
    tmp_= A
    while tmp != tmp_:
      tmp = tmp_
      tmp_ = tmp_ | self. predecessor ( player, tmp_)
    return tmp

  def attractor_gen(self, player, V, A):
    tmp = configBDD.bdd. false
    tmp_ = A & V
    while tmp != tmp_:
      tmp = tmp_
      tmp_ = V & (tmp_ | self. predecessor_gen( player, V, tmp_))    # the use of predecessor_gen and intersection with V are both required!
    return tmp

  def remove (self, A):
    # removing a set of vertices represented by BDD A 
    (self. V, self. even, self. odd) = (self. V & ~A, self. even & ~A, self. odd & ~A)
    A_ = configBDD.bdd. let(self. substitution_list, A)
    self. E = self. E & ~A & ~A_
    p_ = { i : self. p[i] & ~A for i in self. p if self. p[i] & ~A != configBDD.bdd.false}
    self. p = p_
    configBDD.bdd. collect_garbage()
 
  def info (self,debug=False):
    n = len(self.variables)
    print('Number of variables: %d' %(n))
    print('Number of edges: %d' %(configBDD.bdd.count(self.E, 2*n)))
    print('Number of vertices: %d (%d even and %d odd)' %(configBDD.bdd.count(self.V,n), configBDD.bdd.count(self.even,n),configBDD.bdd.count(self.odd,n)))
    if debug: # print all sets
      models = list(configBDD.bdd. pick_iter(self.V, care_vars=self.variables))
      print(' Vertices: %s ' %(models))
      models = list(configBDD.bdd. pick_iter(self.even, care_vars=self.variables))
      print(' Even: %s ' %(models))
      models = list(configBDD.bdd. pick_iter(self.odd, care_vars=self.variables))
      print(' Odd: %s ' %(models))
      for i in self.p:
        models = list(configBDD.bdd. pick_iter(self.p[i], care_vars=self.variables))
        print(' Priority[%d]: %s ' %(i,models))
      models = list(configBDD.bdd. pick_iter(self.E, care_vars=self.variables+self.variables_))
      print(' Edges: %s ' %(models))

 
      
  

  def escape (self, parity, V, U):
    return U & self. predecessor_gen(parity, V, V & ~U)

  def game_size (self):
    return configBDD.bdd.count(self.V,len(self.variables))

  def maximum (self):
    return max(list(self.p.keys())) 

  def PP (self, debug=False):
    (Z0,Z1) = (configBDD.bdd.false,configBDD.bdd.false)
    while self. V != configBDD.bdd.false:
      p = self. maximum()
      (W0,W1) = self. findQuasiDominion(self. V, self. p, p)
      (Z0,Z1) = (Z0 | W0, Z1 | W1)
      if debug: 
        print('Searcher found a dominion of size %d' %(configBDD.bdd.count(W0 | W1, len(self.variables))))
      self. remove(W0 | W1)
    return (Z0,Z1)

  def findQuasiDominion(self, V, priorities, p):
    (parity,opponent) = ('even','odd') if p%2 == 0 else ('odd','even')
    U = priorities[p] & V  
    A = self. attractor_gen(parity, V, U)
    Escape = self. escape(opponent, self.V, A)
    if Escape == configBDD.bdd. false:
      if parity == 'even': return (self.attractor(parity,A), configBDD.bdd.false)
      else: return (configBDD.bdd.false, self.attractor(parity,A))
    else: # subgame is open in the larger game
      priorities_ = {}
      Escape_ = self. escape(opponent,V,A)
      if Escape_ != configBDD.bdd. false:  # A is parity-open
        # compute new priority
        for i in priorities:
          V_i = priorities[i] & ~A
          if i != p and V_i != configBDD.bdd.false: 
            priorities_[i] = V_i
        if p in priorities: 
          priorities_[p] = priorities[p] | A
        else:
          priorities_[p] = A
        p_ = max( [ i for i in list(priorities_.keys()) if i < p and priorities_[i] & V != configBDD.bdd.false ] )
        # computed new priority
        V_ = V & ~A
        return self. findQuasiDominion(V_, priorities_, p_)

      else:  # A is parity-closed
        Opponent = self.odd if parity == 'even' else self.even
        I = self. successor(Opponent & A) & ~A
        p_ = min( [ i for i in list(priorities) if I & priorities[i] != configBDD.bdd.false ] )
        Above = configBDD.bdd.false
        for i in priorities:
          if i >= p_: Above = Above | priorities[i]
        Below = self.V & ~Above
        V_ = configBDD.bdd. false
        for i in self.p: # this guarantees that self.p[i] != configBDD.bdd.false
          p_local = priorities[i] if i in priorities else configBDD.bdd.false
          if i == p_: priorities_[p_] = A | (self.p[p_] & Below) | (p_local & Above)
          else: priorities_[i] = ~A & ((self.p[i] & Below) | (p_local & Above))
          if i <= p_: V_ = V_ | priorities_[i]
       
        return self. findQuasiDominion(V_, priorities_, p_)

  def zielonka (self, debug=False):
    if self. even | self. odd == configBDD.bdd. false:
      return (self. even, self. odd)
    else:
      p = self. maximum ()
      parity = 'even' if p%2 == 0 else 'odd'
      U = self. p[p]
      A = self. attractor (parity, U)
      G = self. duplicate()
      G. remove (A)
      (W0, W1) = G. zielonka ()
      if parity == 'even' and W1 == configBDD.bdd. false:
        return (W0 | A, W1)
      elif parity == 'odd' and W0 == configBDD.bdd. false:
        return (W0, A| W1)
      else:
        (W, opponent) = (W1, 'odd') if parity == 'even' else (W0, 'even')
        B = self. attractor (opponent, W)
        H = self. duplicate()
        H. remove (B)
        (X0, X1) = H. zielonka ()
        if parity == 'even': return (X0, X1 | B) 
        else: return (X0 | B, X1)

  def fixpoint (self, debug=False, optimised = False):

    def Init(i):
      return configBDD.bdd.false if i%2 == 1 else self. V

    def preimage (U):
      # U is a BDD representing a set of vertices
      U_next = configBDD.bdd. let(self. substitution_list, U)
      return configBDD.bdd. quantify(U_next & self. E, self. variables_, forall = False)

    def Diamond(approximation):
      ''' 
        Compute the set { v in V_Even | exists t in V: v --> t && (exists i: t in p[i] && t in X_i) }
        equivalent to:  { v in V_Even | exists i, t in p[i] \cap X_i: v --> t }
        We compute { v in V_Even | exists i, t in p[i] \cap X_i: v --> t } by iterating over i
      ''' 
      U = configBDD.bdd.false
      for i in approximation: U = U | preimage(self.p[i] & approximation[i]) 
      return U & self. even
      
    def Box(approximation):
      ''' 
        Compute the set { v in V_Odd | forall t in V: v --> t => (forall i: t in p[i] => t in X_i) }
        equivalent to:  { v in V_Odd | not (exists t in V: v --> t && (exists i: t in p[i] && t not in X_i) ) }
        equivalent to:  { v in V_Odd | not (exists i, t in p[i]\X_i, i: v --> t ) }
        equivalent to: V_Odd \ { v in V_Odd | exists i, t in p[i]\X_i, i: v --> t }
        We first compute { v in V_Odd | exists i, t in p[i]\X_i, i: v --> t } by iterating over i
      '''
      U = configBDD.bdd.false
      for i in approximation: U = U | preimage(self.p[i] & ~approximation[i])
      return self.odd & ~U 

    (d_min,d_max) = (min(list(self.p)), max(list(self.p)))
    (Approx,Approx_) = ({},{})
    for i in list(self. p): Approx[i] = Init(i)  # initialisation
    while True:
      Approx_[d_min] = Approx[d_min]
      Approx[d_min] = Diamond(Approx) | Box(Approx)
      (i,j) = (d_min,d_min)
      if optimised:
        j = i+1
        while j < d_max and Approx[j] == Approx[j-1]:
          j = j+1
      while Approx[i] == Approx_[i] and i < d_max:
        i = i +1
        Approx_[i] = Approx[i]
        Approx[i] = Approx[i-1]
        if not(optimised) or (i%2 == j%2): Approx[i-1] = Init(i-1)
      if (i == d_max) and (Approx[i] == Approx_[i]): 
        break
    return (Approx[d_max], self.V & ~Approx[d_max])

